require "numru/gphys"
require "numru/dcl"
require "numru/misc"

module NumRu

  # add methods to GPhys
  class GPhys
    def first2D
      raise "rank less than 2" if rank < 2
      self[true,true,*([0]*(rank-2))]
    end
    def first1D
      raise "rank less than 1" if rank < 1
      self[true,*([0]*(rank-1))]
    end
  end

###### DCLExt is to be moved into the RubyDCL distribution ########

=begin
=module NumRu::DCLExt

Collection of various compound DCL functions for convenience.

==Index
MATH1
* ((<glpack>))
GRPH1
* ((<sgpack>))
* ((<slpack>))
* ((<swpack>))
GRPH2
* ((<uzpack>))
* ((<ulpack>))
* ((<ucpack>))
* ((<uupack>))
* ((<uspack>))
* ((<udpack>))
* ((<uepack>))
* ((<ugpack>))
* ((<umpack>))

==Module Functions

===glpack
---gl_set_params(hash)
    Calls (({DCL.glpset})) multiple times (for each key and val of (({hash}))).

    ARGUMENTS
    * hash (Hash) : combinations of parameter names and values for udpset

    RETURN VALUE
    * a Hash containing the parameter names and their old values that were 
      replaced.

    EXAMPLES
    * You can modify parameters temporarilly as follows.

        before = DCLExt.gl_set_params('lmiss'=>true,'rmiss'=>9999.0)
        ....
        DCLExt.gl_set_params(before)     # reset the change

===sgpack
---sg_set_params(hash)
    Calls (({DCL.sgpset})) multiple times (for each key and val of (({hash}))).

    See ((<gl_set_params>)) for usage.

===slpack
---sl_set_params(hash)
    Calls (({DCL.slpset})) multiple times (for each key and val of (({hash}))).

    See ((<gl_set_params>)) for usage.

===swpack
---sw_set_params(hash)
    Calls (({DCL.swpset})) multiple times (for each key and val of (({hash}))).

    See ((<gl_set_params>)) for usage.

===uzpack
---uz_set_params(hash)
    Calls (({DCL.uzpset})) multiple times (for each key and val of (({hash}))).

    See ((<gl_set_params>)) for usage.

===ulpack
---ul_set_params(hash)
    Calls (({DCL.ulpset})) multiple times (for each key and val of (({hash}))).

    See ((<gl_set_params>)) for usage.

===ucpack
---uc_set_params(hash)
    Calls (({DCL.ucpset})) multiple times (for each key and val of (({hash}))).

    See ((<gl_set_params>)) for usage.

===uupack
---uu_set_params(hash)
    Calls (({DCL.uupset})) multiple times (for each key and val of (({hash}))).

    See ((<gl_set_params>)) for usage.

===uspack
---us_set_params(hash)
    Calls (({DCL.uspset})) multiple times (for each key and val of (({hash}))).

    See ((<gl_set_params>)) for usage.

===udpack
---ud_set_params(hash)

    Calls (({DCL.udpset})) multiple times (for each key and val of (({hash}))).

    ARGUMENTS
    * hash (Hash) : combinations of parameter names and values for udpset

    RETURN VALUE
    * a Hash containing the parameter names and their old values that were 
      replaced.

    EXAMPLES
    * You can modify parameters temporarilly as follows.

        before = DCLExt.ud_set_params('indxmj'=>4,'lmsg'=>false)
        DCL.udcntz(data)
        DCLExt.ud_set_params(before)     # reset the change

---ud_set_linear_levs(v, options)

    Set contour levels with a constant interval

    ARGUMENTS
    * v : Data values to be fed to udcnt[rz]
    * options (Hash) : option specification by keys and values. Avaliable
      options are 
          name   default value   description
          'min'      nil         minimum contour value
          'max'      nil         maximum contour value
          'nlev'     nil         number of levels
          'interval' nil         contour interval
          'nozero'   nil         delete zero contour
          'coloring', false,     set color contours with ud_coloring
          'clr_min', 13,         (if coloring) minimum color id
          'clr_max', 100,        (if coloring) maximum color id
      Here, (({interval})) has a higher precedence over (({nlev})).
      Since all the default values are nil, only those explicity specified
      are interpreted. If no option is provided, the levels generated will
      the default ones set by udcnt[rz] without any level specification.

---ud_set_linear_lev_params(hash)

    Changes the default value of the options of ((<ud_set_linear_levs>))

    ARGUMENTS
    * options (Hash) : see ((<ud_set_linear_levs>))

    RETURN VALUE
    * a Hash containing the values replaced (keys are the option names)

---ud_next_linear_lev_params(hash)

    Similar to ((<ud_set_linear_lev_params>)), but the setting is effective
    only for the next call of ((<ud_set_linear_levs>)).

---ud_set_levs(levels,index=nil,line_type=nil,label=nil,label_height=nil)

    Set contours of at specified levels.

    Normally you do not have to specify (({label})) and (({label_height})).

    It calls DCL.udsclv for each level. So the arguments are basically
    the same as DCL.udsclv, but only levels are mandatory here.

    ARGUMENTS
    * levels (Array, NArray, or Numeric) : contour levels to be set.
      If Numeric, a single level is set.
    * index (Array of integers, Integer, or nil) : 
      index(es) of the contours. If Array (or NArray), 
      the length must agree with that of (({levels})).
      If it is a single Integer, all the contour will have the same index. 
     If nil, the value of 'indxmn' is used.
    * line_type (Array of integers, Integer, or nil) : 
      line type(s) of the contours. If Array (or NArray), 
      the length must agree with that of (({levels})).
      If it is a single Integer, all the contour will have the same type. 
      If nil, set to be 1.
    * label (Array of String, String, true, false, nil) :
      Label(s) of the conoutrs. If Array (or NArray), 
      the length must agree with that of (({levels})).
      If  it is a single Strin, all the contour will have the same label. 
      If true, all the contours will have the labels representing the levels.
      If false, no label will be drawn (set to "").
      If nil, same as true for the contours whose index is equal to "INDXMJ",
      and same as false otherwise.
    * label_height (Array of Numeric, Numeric, or nil) :
      Heigh of Labels. Normally you do not have to use this.
      If Array (or NArray), 
      the length must agree with that of (({levels})).
      If nil, the default value ("RSIZEL") is used for non-empty labels.
      If a single Numeric, the same value is used for all the countours.
      Note that it is recommended to not to use this parameter but
      use DCL.udpset('RZISEL'. label_height), since a positive value
      here always means to draw labels even when the label is empty.

    RETURN VALUE
    * nil

---ud_add_levs(levels,index=nil,line_type=nil,label=nil,label_height=nil)

    Same as ((<ud_set_levs>)), but does not clear the contour levels that have
    been set.

===uepack
---ue_set_params(hash)
    Calls (({DCL.uepset})) multiple times (for each key and val of (({hash}))).

    See ((<gl_set_params>)) for usage.

===ugpack
---ug_set_params(hash)
    Calls (({DCL.ugpset})) multiple times (for each key and val of (({hash}))).
    See ((<gl_set_params>)) for usage.

===umpack
---um_set_params(hash)
    Calls (({DCL.umpset})) multiple times (for each key and val of (({hash}))).

    See ((<gl_set_params>)) for usage.

=end

  module DCLExt
    # to be included in the RubyDCL distribution

    module_function

    #<<< for many packages >>>

    %w!gl sg sl sw uz ul uc uu us ud ue ug um!.each do |pkg|
      eval <<-EOS
        def #{pkg}_set_params(hash)
	  before = Hash.new
	  hash.each{|k,v|
	    before[k]=DCL.#{pkg}pget(k)
	    DCL.#{pkg}pset(k,v)
	  }
	  before
	end
      EOS
    end

    #<<< udpack >>>

    def ud_coloring(clr_min=13, clr_max=100)
      # change the colors of existing contours to make a gradation
      # (rainbow colors with the default color map).
      nlev = DCL.udqcln
      cont_params = Array.new
      for i in 1..nlev
	cont_params.push( DCL.udqclv(i) )   # => [zlev,indx,ityp,clv,hl]
      end
      DCL.udiclv     # clear the contours

      colors = clr_min + 
	       NArray.int(nlev).indgen! * (clr_max-clr_min) / nlev

      cont_params.sort!
      for i in 0...nlev
	cont_params[i][1] += colors[i]*10   # indx += colors[i]*10
        DCL.udsclv(*cont_params[i])
      end
    end

    @@ud_linear_lev_params = Misc::KeywordOpt.new(
      ['min',      nil,       'minimum contour value'],
      ['max',      nil,       'maximum contour value'],
      ['nlev',     nil,       'number of levels'],
      ['interval', nil,       'contour interval'],
      ['nozero',   nil,       'delete zero contour'],
      ['coloring', false,     'set color contours with ud_coloring'],
      ['clr_min',  13,        '(if coloring) minimum color id'],
      ['clr_max',  100,       '(if coloring) maximum color id']
    )

    def ud_set_linear_lev_params(hash)
      @@ud_linear_lev_params.set(hash)
    end

    @@ud_next_linear_lev_params = nil
    def ud_next_linear_lev_params(hash)  
      # only for the next ud_set_linear_levs
      @@ud_next_linear_lev_params = hash    # flushed by ud_set_linear_levs
    end

    def ud_set_linear_levs(v, options=nil)
      if @@ud_next_linear_lev_params 
	options = ( options ? @@ud_next_linear_lev_params.update( options ) :
	                    @@ud_next_linear_lev_params )
	@@ud_next_linear_lev_params = nil
      end
      par = @@ud_linear_lev_params.interpret(options)
      min = par['min']
      max = par['max']
      nlev = par['nlev']
      interval = par['interval']
      nozero = par['nozero']
      if interval
	dx = interval
      elsif nlev
	dx = -nlev
      else
	dx = 0
      end
      if min || max
 	min = v.min if !min
 	max = v.max if !max
	DCL.udgcla(min, max, dx)
      else
	DCL.udgclb(v, dx)
      end
      if nozero
	DCL.uddclv(0.0)
      end
      if par['coloring']
	ud_coloring( par['clr_min'], par['clr_max'] )
      end
    end

    def ud_set_levs(*args)
      DCL.udiclv
      ud_add_levs(*args)
    end

    def ud_add_levs(levels,index=nil,line_type=nil,label=nil,label_height=nil)

      # < check levels >
      case levels
      when Array, NArray
	# This is expected. Nothing to do.
      when Numric
	levels = [levels]
      else
	raise ArgumentError, "invalid level specification (#{levels})"
      end

      nlev = levels.length

      # < index >
      case index
      when Array, NArray
	if index.length != nlev
	  raise ArgumentError, "lengths mismatch btwn levels(#{nlev})" +
                " and index (#{index.length})"
	end
      when Numeric
	index = [index]*nlev
      when nil
	index = [DCL.udpget('indxmn')]*nlev
      else
	raise ArgumentError, "unsupported index type (#{index.class})"
      end

      # < line_type >
      case line_type
      when Array, NArray
	if line_type.length != nlev
	  raise ArgumentError, "lengths mismatch btwn levels(#{nlev})" +
                " and line_type(#{line_type.length})"
	end
      when Numeric
	line_type = [line_type]*nlev
      when nil
	line_type = [1]*nlev
      else
	raise ArgumentError, "unsupported index type (#{index.class})"
      end

      # < label >
      case label
      when Array, NArray
	if label.length != nlev
	  raise ArgumentError, "lengths mismatch btwn levels(#{nlev})" +
                " and label(#{label.length})"
	end
      when String
	label = [label]*nlev
      when false
	label = [""]*nlev
      when true
	label = (0...nlev).collect{|i| 
	    DCL.udlabl(levels[i])
	}
      when nil
	indxmj = DCL.udpget('indxmj')
	label = (0...nlev).collect{|i| 
	  if index[i]==indxmj
	    DCL.udlabl(levels[i])
	  else
	    ""
	  end
	}
      else
	raise ArgumentError, "unsupported index type (#{index.class})"
      end

      # < label_height >
      case label_height
      when Array, NArray
	if label_height.length != nlev
	  raise ArgumentError, "lengths mismatch btwn levels(#{nlev})" +
                " and label_height(#{label_height.length})"
	end
      when Numeric
	label_height = [label_height]*nlev
      when nil
	label_height = label.collect{|lv| lv=="" ? 0.0 : DCL.udpget('rsizel')}
      else
	raise ArgumentError, "unsupported index type (#{index.class})"
      end

      # < set levels >

      for i in 0...nlev
	DCL.udsclv(levels[i],index[i],line_type[i],label[i],label_height[i])
      end
      nil
    end

    #<<< uepack >>>

    @@ue_linear_lev_params = Misc::KeywordOpt.new(
      ['min',      nil,       'minimum tone value'],
      ['max',      nil,       'maximum tone value'],
      ['nlev',     nil,       'number of levels'],
      ['interval', nil,       'tone interval'] # ,
      ##['nozero',   nil,       'delete zero tone level'],
      ##['clr_min',  13,        '(if coloring) minimum color id'],
      ##['clr_max',  100,       '(if coloring) maximum color id']
    )

    def ue_set_linear_lev_params(hash)
      @@ue_linear_lev_params.set(hash)
    end

    @@ue_next_linear_lev_params = nil
    def ue_next_linear_lev_params(hash)  
      # only for the next ue_set_linear_levs
      @@ue_next_linear_lev_params = hash    # flushed by ue_set_linear_levs
    end

    def ue_set_linear_levs(v, options=nil)
      if @@ue_next_linear_lev_params 
	options = ( options ? @@ue_next_linear_lev_params.update( options ) :
	                    @@ue_next_linear_lev_params )
	@@ue_next_linear_lev_params = nil
      end
      par = @@ue_linear_lev_params.interpret(options)
      min = par['min']
      max = par['max']
      nlev = par['nlev']
      interval = par['interval']
      nozero = par['nozero']
      if interval
	dx = interval
      elsif nlev
	dx = -nlev
      else
	dx = 0
      end
      if min || max
 	min = v.min if !min
 	max = v.max if !max
	DCL.uegtla(min, max, dx)
      else
	DCL.uegtlb(v, dx)
      end
    end

  end

############################################################

  module GGraph

    module_function

    @@fig_params = Misc::KeywordOpt.new(
      ['viewport',[0.2,0.8,0.2,0.8], '[vxmin, vxmax, vymin, vymax]'],
      ['itr',     1,     'coordinate transformation number']
    )

    def set_fig_params(hash)
      @@fig_params.set(hash)
    end

    def fig(window, hash_fig_params)
      figpar = @@fig_params.interpret(hash_fig_params)
      DCL.grfig
      DCL.grsvpt(*figpar['viewport'])
      DCL.grswnd(*window)
      DCL.grstrn(figpar['itr'])
      DCL.grstrf
    end

    def axes(xax=nil, yax=nil, cxside='tb', cyside='lr')
      if xax
	xai = _get_axinfo(xax) 
	DCL.uscset('cxttl', xai['title'])
	DCL.uscset('cxunit',xai['units'])
	cxside.scan('.').each{|s| DCL.usxaxs(s)}   # for each char in cxside
      end
      if yax
	yai = _get_axinfo(yax)
	DCL.uscset('cyttl', yai['title'])
	DCL.uscset('cyunit',yai['units'])
	cyside.scan('.').each{|s| DCL.usyaxs(s)}   # for each char in cyside
      end
    end

    def _get_axinfo(vary)
      raise ArgumentError,"not a VArray" if !vary.is_a?(VArray)
      raise "Not 1D" if vary.rank!=1
      return { #'val'=>vary.val , 
	      'title'=>(vary.attr['long_name'] || vary.name).gsub('_','' ) ,
	      'units'=>(vary.attr['units'] || '').gsub('_',' ')
             }
    end

    def contour(gphys, newframe=true, options=nil)
      if newframe!=true && newframe!=false
	raise ArgumentError, "2nd arg (newframe) must be true or false"
      end
      if ! defined?(@@contour_options)
	@@contour_options = Misc::KeywordOpt.new(
          ['fig_params', nil, 'if non-nil, a hash specifying the parameters of fig'],
          ['set_levs',  true, 'true: set levels internally, nil/false: use those that have been set'],
          ['linear_levs', nil, 'if non-nil, a hash specifying options of DCL_Ext::ud_set_linear_levs'],
          ['ud_params', nil, 'if non-nil, a hash specifying parameters for DCL_Ext::ud_set_params (temporary changed and set back)']
        )
      end
      options = @@contour_options.interpret(options)
      gp = gphys.first2D
      if newframe
	xax = gp.axis(0).pos
	yax = gp.axis(1).pos
	window=[xax.val.min, xax.val.max, yax.val.min, yax.val.max]
	DCL.grfrm
	fig(window, options['fig_params'])
	axes(xax, yax)
      end
      if options['ud_params']
	before = DCLExt.ud_set_params(options['ud_params'])
      end
      if options['set_levs'] || options['linear_levs']
	if options['linear_levs']==true; options['linear_levs']=nil; end
        DCLExt.ud_set_linear_levs(gp.data.val, options['linear_levs'])
      end
      DCL.udcntz(gp.data.val)
      if options['ud_params']
	DCLExt.ud_set_params(before)
      end
    end

    def tone(gphys, newframe=true, options=nil)
      if newframe!=true && newframe!=false
	raise ArgumentError, "2nd arg (newframe) must be true or false"
      end
      if ! defined?(@@tone_options)
	@@tone_options = Misc::KeywordOpt.new(
          ['fig_params', nil, 'if non-nil, a hash specifying the parameters of fig'],
          ['set_levs',  false, 'true: set levels internally, nil/false: use those that have been set'],
          ['linear_levs', nil, 'if non-nil, a hash specifying options of DCL_Ext::ud_set_linear_levs'],
          ['ue_params', nil, 'if non-nil, a hash specifying parameters for DCL_Ext::ue_set_params (temporary changed and set back)']
	)
      end
      options = @@tone_options.interpret(options)

      gp = gphys.first2D
      if newframe
	xax = gp.axis(0).pos
	yax = gp.axis(1).pos
	window=[xax.val.min, xax.val.max, yax.val.min, yax.val.max]
	DCL.grfrm
	fig(window, options['fig_params'])
      end
      if options['ue_params']
	before = DCLExt.ue_set_params(options['ue_params'])
      end
      if options['set_levs'] || options['linear_levs']
	if options['linear_levs']==true; options['linear_levs']=nil; end
        DCLExt.ue_set_linear_levs(gp.data.val, options['linear_levs'])
	DCL.uetone(gp.data.val)
	DCL.ueitlv
      else
	DCL.uetone(gp.data.val)
      end
      if newframe
	axes(xax, yax)  # axes should be drawn after uetone
      end
      if options['ue_params']
	DCLExt.ue_set_params(before)
      end
    end

  end

end

if $0 == __FILE__
  include NumRu
  file = '../../testdata/T.jan.nc'
  gphys = GPhys::NetCDF_IO.open(file, 'T')
  DCL.gropn(1)
  GGraph.contour(gphys)
  ##DCLExt.ud_set_params('icycle'=>3)   # parmanent change
  GGraph.contour(gphys,true, 'linear_levs'=>{'min'=>0, 'color'=>true}, 
		 'ud_params'=>{'icycle'=>3})
  GGraph.set_fig_params('viewport'=>[0.15,0.85,0.3,0.7])
  GGraph.contour(gphys,true, 'linear_levs'=>{'nozero'=>true})
  GGraph.contour(gphys,true, 'linear_levs'=>{'min'=>10, 'int'=>3})
  GGraph.contour(gphys,false, 'linear_levs'=>{'max'=>-10, 'int'=>3},
		              'ud_params'=>{'lmsg'=>false})
  DCLExt.ud_set_levs([-10,-5,0,5,10],[3,1,3,1,3],3)
  GGraph.contour(gphys, true, 'set_levs'=> false)
  DCLExt.ud_set_linear_lev_params('nlev'=>24)
  DCLExt.ud_next_linear_lev_params('coloring'=>true)   # only for the next
  GGraph.contour(gphys)

#  DCL.sgpset('lsoftf',true)
  GGraph.tone(gphys, true, 'linear_levs'=>{'min'=>0})
  GGraph.tone(gphys)
  GGraph.contour(gphys, false)
  # GGraph.set_fig_params('itr'=>3)
  GGraph.contour(gphys[1..-1,true,true], true, 'fig_param'=>{'itr'=>3})
  DCL.grcls
end
